"""Implementation of pairs for Python."""

from types import ListType as __ListType
from types import TupleType as __TupleType
from symbol import Symbol
from error import SchemeError
from operator import add as __add

def isNull(x):
    return x == []


def schemeCons(head, rest):
    """Returns the concatentation of head with rest.

    Note: If rest isn't a list, we use a peculiar representation of a
    pair:

        ([head], [rest]).

    The reason for this is to allow set-car! and set-cdr! to work if we decide
    to implement it.
    """
    return ([head], [rest])


def schemeCar(L):
    if type(L) not in (__ListType, __TupleType):
        raise SchemeError, "CAR --- argument is not a pair."
    if len(L) == 0:
        raise SchemeError, "CAR --- cannot take CAR of empty list."
    if type(L) == __ListType: return L[0]
    else: return L[0][0]


def schemeCdr(L):
    if type(L) not in (__ListType, __TupleType):
        raise SchemeError, "CDR --- argument is not a pair."
    if len(L) == 0:
        raise SchemeError, "CDR --- cannot take CDR of empty list."
    if type(L) == __ListType: return L[1:]
    else: return L[1][0]


def schemeSetCarBang(pair, element):
    if type(pair) != __TupleType:
        raise SchemeError, "SET-CAR! --- cannot set car of non-pair."
    pair[0][0] = element
    return Symbol("ok")


def schemeSetCdrBang(pair, element):
    if type(pair) != __TupleType:
        raise SchemeError, "SET-CAR! --- cannot set car of non-pair."
    pair[1][0] = element
    return Symbol("ok")


def isPair(L):
    return type(L) in (__ListType, __TupleType)


def isDottedPair(L):
    if not isPair(L): return 0
    elif isPair(schemeCdr(L)): return 0
    else: return 1


def schemePairQuestion(L):
    return schemeBooleanize(isPair(L))


def schemeListQuestion(x):
    if not isPair(schemeCar(L)):
        return false
    if isNull(L):
        return true
    return schemeListQuestion(schemeCdr(L))


def pairsToPythonList(pair):
    elements = []
    while not isNull(pair):
        elements.append(schemeCar(pair))
        pair = schemeCdr(pair)
    return elements


def all(f, l):
    """Returns true of all elements return a true value on f."""
    for element in l:
        if f(l): continue
        else: return 0
    return 1



def appendTwo(front, back):
    if isNull(front): return back
    return schemeCons(schemeCar(front),
                      appendTwo(schemeCdr(front), back))
    
def schemeAppend(*lists):
    if len(lists) == 0: return []
    if all(lambda x: type(x) == __ListType, lists):
        return reduce(__add, lists)
    if not all(isPair, lists[:-1]):
        raise SchemeError, "APPEND: arguments to APPEND must be pairs."

    appended_list = lists[-1]
    for i in range(len(lists)-2, -1, -1):
        next_list = lists[i]
        appended_list = appendTwo(next_list, appended_list)
    return appended_list


def schemeList(*elements):
    result = list(elements)
    result.reverse()
    return reduce(lambda x, y: schemeCons(y, x), [[]] + result)
